@@ -36,10 +36,11 @@ static void hv_store_or_croak(HV *hv, const char *key, SV *val) {
   croak("Can't store value");
 }
 
-static void av_store_or_croak(AV *hv, int key, SV *val) {
-  if (av_store(hv, key, val)) return;
-  SvREFCNT_dec(val);
-  croak("Can't store value");
+static inline SV *av_store_or_croak(AV *av, int key, SV *val) {
+  SV **stored = av_fetch(av, key, 1);;
+  if(!stored) croak("Can't store value");
+  SvSetSV(*stored, val);
+  return *stored;
 }
 
 // for tests - return a shape of registered attrs
@@ -58,21 +59,21 @@ static SV *ecg_to_array(ClassGen *gen) {
     char *rtype = "";
 
     switch (attr->rtype) {
-    case ECG_RELAXED:
-      rtype = "relaxed";
-      break;
-    case ECG_DEFAULT:
-      rtype = "default";
-      break;
-    case ECG_DEFAULT_CODE:
-      rtype = "default_code";
-      break;
-    case ECG_REQUIRED:
-      rtype = "required";
-      break;
-    case ECG_LAZY:
-      rtype = "lazy";
-      break;
+      case ECG_RELAXED:
+        rtype = "relaxed";
+        break;
+      case ECG_DEFAULT:
+        rtype = "default";
+        break;
+      case ECG_DEFAULT_CODE:
+        rtype = "default_code";
+        break;
+      case ECG_REQUIRED:
+        rtype = "required";
+        break;
+      case ECG_LAZY:
+        rtype = "lazy";
+        break;
     }; // let compiler notify us about unused, don't use default
 
     hv_store_or_croak(slot, "rtype", newSVpv(rtype, 0));
@@ -148,15 +149,16 @@ static ClassGen *find_gen(SV *cv) {
   return INT2PTR(ClassGen *, SvIV((SV *)SvRV(mg->mg_obj)));
 }
 
-AV *init_storage(SV *obj) {
+static inline AV *init_storage(SV *obj, SSize_t init_count ) {
   if (!SvROK(obj)) croak("Not a ref");
   AV *array = newAV();
+  if(init_count > -1) av_extend(array, init_count);
   sv_magicext(SvRV(obj), (SV *)array, PERL_MAGIC_ext, &STORAGE_TBL, NULL, 0);
   SvREFCNT_dec(array);
   return array;
 }
 
-AV *get_storage(SV *obj) {
+static inline AV *get_storage(SV *obj) {
   if (!SvROK(obj)) croak("Not a ref");
   MAGIC *mg = mg_findext(SvRV(obj), PERL_MAGIC_ext, &STORAGE_TBL);
   if (!mg) croak("No magic, forgot to init object?");
@@ -220,31 +222,40 @@ void invoke_check(SV *cv, SV *value, char *name) {
 
   if (!SvTRUE(ok))
     croak("Bad value \"%s\" for attribute \"%s\": %s", SvPV_nolen(value), name,
-          SvTRUE(msg) ? SvPV_nolen(msg) : "");
+        SvTRUE(msg) ? SvPV_nolen(msg) : "");
 };
 
 SV *invoke_lazy(SV *self, SV *cv) {
   dSP;
 
+  ENTER;
+  SAVETMPS;
+
   PUSHMARK(SP);
   PUSHs(self);
   PUTBACK;
   int count = call_sv(cv, G_SCALAR);
   if (count != 1) croak("bad count");
   SPAGAIN;
-  SV *result = newSVsv(POPs);
-  PUTBACK;
+  SV *res = POPs;
+  SV *result = newSVsv(res);
 
-  return result;
+  PUTBACK;
+  FREETMPS;
+  LEAVE;
+
+
+  return sv_2mortal(result);
 }
 
-static void do_init(SV *cv, SV *class, SV *obj, SV *args[], int count) {
+static inline void do_init(SV *cv, SV *class, SV *obj, SV *args[], int count) {
   int offset = args_to_uniq(args, count);
 
-  AV *storage = init_storage(obj);
-
   ClassGen *gen = find_gen(cv);
   if (!gen) goto IN_GD; // global destruction?
+
+  AV *storage = init_storage(obj, gen->size-1);
+
   // // iterate passed opts
   for (int i = offset; i < count; i += 2) {
     int index = name2index(gen, SvPV_nolen(args[i]));
@@ -252,7 +263,7 @@ static void do_init(SV *cv, SV *class, SV *obj, SV *args[], int count) {
     char *key = SvPV_nolen(args[i]);
     SV *check = gen->attrs[index]->check;
     if (check) invoke_check(check, val, key);
-    av_store_or_croak(storage, index, SvREFCNT_inc(val));
+    av_store_or_croak(storage, index, val);
   }
 
   // iterate known attributes
@@ -262,10 +273,10 @@ static void do_init(SV *cv, SV *class, SV *obj, SV *args[], int count) {
     if (attr->rtype == ECG_REQUIRED) {
       croak("Attribute \"%s\" is required", attr->name);
     } else if (attr->rtype == ECG_DEFAULT) {
-      av_store_or_croak(storage, i, newSVsv(attr->rvalue));
+      av_store_or_croak(storage, i, attr->rvalue);
     } else if (attr->rtype == ECG_DEFAULT_CODE) {
-      SV *val = invoke_lazy(class, attr->rvalue);
-      av_store_or_croak(storage, i, val);
+      SV *res =  invoke_lazy(class, attr->rvalue);
+      av_store_or_croak(storage, i, res);
     }
   }
 
@@ -297,8 +308,8 @@ static void xs_attr(SV *cv) {
       ClassGenAttr *attr = gen->attrs[index];
 
       if (attr->rtype == ECG_LAZY) {
-        result = invoke_lazy(self, attr->rvalue);
-        av_store_or_croak(storage, index, result);
+        SV *tmp = invoke_lazy(self, attr->rvalue);
+        result = av_store_or_croak(storage, index, tmp);
       } else {
         result = &PL_sv_undef;
       }
@@ -312,14 +323,13 @@ static void xs_attr(SV *cv) {
     SV *value = ST(1);
     if (attr->check) invoke_check(attr->check, value, attr->name);
     av_store_or_croak(storage, index, value);
-    SvREFCNT_inc(value);
     result = self;
   }
 
   ST(0) = result;
   XSRETURN(1);
 
-// global destruction?
+  // global destruction?
 IN_GD_GET:
   ST(0) = &PL_sv_undef;
   XSRETURN(1);
